/**
 * @file decoder.c
 * @brief Source of the decoding functionality, handles rotation as well as position decoding
 * @author Lukas Hostettler
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#ifdef __cplusplus
extern "C" {
#endif

#include "compiler.h"
#include "config.h"

#include "decoder.h"

#pragma GCC push_options
#pragma GCC optimize ("O1")

#define INVALID_COEFF -1
#define MAIN_SEQUENCE_LENGTH 63
#define PRIMARY_SEQUENCE_LENGTH 8

/********************************
 * Global variable declarations *
 ********************************/
///@var Main number sequence expressed for convolutions
const static char mainNumberSequenceForCorr[] = {
    -1,-1,-1,-1,-1,-1,1,-1,-1,1,1,1,1,1,-1,1,-1,-1,1,-1,-1,-1,-1,1,1,1,-1,1,1,1,-1,-1,1,-1,1,-1,1,-1,-1,-1,1,-1,1,1,-1,
    1,1,-1,-1,1,1,-1,1,-1,1,1,1,1,-1,-1,-1,1,1,-1,-1,-1,-1,-1,-1,1,-1,-1,1,1,1,1,1,-1,1,-1,-1,1,-1,-1,-1,-1,1,1,1,-1,1,
    1,1,-1,-1,1,-1,1,-1,1,-1,-1,-1,1,-1,1,1,-1,1,1,-1,-1,1,1,-1,1,-1,1,1,1,1,-1,-1,-1,1,1
}; //Main number sequence of length 63 is repeated for convolution; we can actually remove the last 63 - 8 = 55 entries

const static unsigned char cSeq0[] = {
    0,0,0,0,0,1,0,0,0,0,2,0,1,0,0,1,0,1,0,0,2,0,0,0,1,1,0,0,0,1,2,0,0,1,0,2,0,0,2,0,2,0,1,1,0,1,0,1,1,0,2,0,1,2,0,1,0,
    1,2,0,2,1,0,0,1,1,1,0,1,1,1,1,0,2,1,0,1,0,2,1,1,0,0,1,2,1,0,1,1,2,0,0,0,2,1,0,2,0,2,1,1,1,0,0,2,1,2,0,1,1,1,2,0,2,
    0,0,1,1,2,1,0,0,0,2,2,0,1,0,2,2,0,0,1,2,2,0,2,0,2,2,1,0,1,2,1,2,1,0,2,1,2,1,1,0,2,2,1,2,1,2,0,2,2,0,2,2,2,0,1,1,2,
    2,1,1,0,1,2,2,2,2,1,2,0,0,2,2,1,1,2,1,2,2,1,0,2,2,2,2,2,0,2,1,2,2,2,1,1,1,2,1,1,2,0,1,2,2,1,2,2,0,1,2,1,1,1,1,2,2,
    2,0,0,2,1,1,2,2/*make it cyclic*/,0,0,0,0,0,1,0
};

const static unsigned char cSeq1[] = {
    0,0,0,0,0,1,0,0,0,0,2,0,1,0,0,1,0,1,0,1,1,0,0,0,1,1,1,1,0,0,1,1,0,1,0,0,2,0,0,0,1,2,0,1,0,1,2,1,0,0,0,2,1,1,1,0,1,
    1,1,0,2,1,0,0,1,2,1,2,1,0,1,0,2,0,1,1,0,2,0,0,1,0,2,1,2,0,0,0,2,2,0,0,1,1,2,0,2,0,0,2,0,2,0,1,2,0,0,2,2,1,1,0,0,2,
    1,0,1,1,2,1,0,2,0,2,2,1,0,0,2,2,2,1,0,1,2,2,0,0,2,1,2,2,1,1,1,1,1,2,0,0,1,2,2,1,2,0,1,1,1,2,1,1,2,0,1,2,1,1,1,2,2,
    0,2,2,0,1,1,2,2,2,2,1,2,1,2,2,0,1,2,2,2,0,2,0,2,1,1,2,2,1,0,2,2,0,2,1,0,2,1,1,0,2,2,2,2,0,1,0,2,2,1,2,2,2,1,1,2,1,
    2,0,2,2,2/*make it cyclic*/,0,0,0,0,0,1,0
};

const static unsigned char cSeq2[] = {
    0,0,0,0,0,1,0,0,1,1,0,0,0,1,1,1,1,0,0,1,0,1,0,1,1,0,1,1,1,0,1/*make it cyclic*/,0,0,0,0,0,1,0
};

const static char unsigned cSeq3[] = {
    0,0,0,0,0,1,0,2,0,0,0,0,2,0,0,2,0,1,0,0,0,1,1,2,0,0,0,1,2,0,0,2,1,0,0,0,2,1,1,2,0,1,0,1,0,0,1,2,1,0,0,1,0,0,2,2,0,
    0,0,2,2,1,0,2,0,1,1,0,0,1,1,1,0,1,0,1,1,0,1,2,0,1,1,1,1,0,0,2,0,2,0,1,2,0,2,2,0,1,0,2,1,0,1,2,1,1,0,1,1,1,2,2,0,0,
    1,0,1,2,2,2,0,0,2,2,2,0,1,2,1,2,0,2,0,0,1,2,2,0,1,1,2,1,0,2,1,1,0,2,0,2,1,2,0,0,1,1,0,2,1,2,1,0,1,0,2,2,0,2,1,0,2,
    2,1,1,1,2,0,2,1,1,1,0,2,2,2,2,0,2,0,2,2,1,2,1,1,1,1,2,1,2,1,2,2,2,1,0,0,2,1,2,2,1,0,1,1,2,2,1,1,2,1,2,2,2,2,1,2,0,
    1,2,2,1,2,2,0,2,2,2,1,1,1/*make it cyclic*/,0,0,0,0,0,1,0
};

const static unsigned char* cSequences[] = {cSeq0, cSeq1, cSeq2, cSeq3};                                         ///@var Array of the secondary number sequences
const static unsigned char cSequenceLengths[] = {236, 233, 31, 241};                                             ///@var Secondary number sequence lengths
const static unsigned int otherCSequenceLengthsProduct[] = {233*31*241, 236*31*241, 236*233*241, 236*233*31};    ///@var Product of the other secondary number sequence lengths at each index
const static unsigned int cSequenceLengthsProduct = 236*233*31*241;                                              ///@var Product of all secondary number sequence lengths
const static unsigned int cSequenceLengthsMulInvs[] = {135, 145, 17, 62};                                        ///@var Modular multiplicative inverses of otherCSequenceLengthsProduct entries (modulo cSequenceLengths entries at each index)
const static unsigned char cSequenceShifts[] = {46, 42, 15, 51};
const static unsigned int cSequenceMultipliers[] = {1, 3, 9, 18};

///@var Fast LookupTable for 8-bit sequence in order to avoid correlations
static int mainLookUp[] = {
    -1,-1,0,-1,1,-1,-1,-1,-1,2,-1,-1,-1,-1,19,-1,-1,-1,-1,3,-1,-1,37,-1,58,-1,-1,-1,-1,20,-1,-1,-1,16,-1,-1,-1,-1,-1,4,
    -1,-1,30,-1,-1,38,-1,-1,59,-1,-1,-1,-1,47,-1,-1,-1,-1,-1,21,-1,-1,7,-1,-1,-1,-1,17,-1,35,-1,-1,14,-1,-1,-1,-1,-1,
    -1,5,-1,33,-1,-1,31,-1,-1,-1,-1,-1,-1,39,-1,-1,51,-1,60,-1,-1,-1,-1,-1,44,-1,-1,-1,-1,48,41,-1,-1,-1,-1,-1,26,-1,
    -1,-1,-1,22,53,-1,-1,-1,-1,8,-1,-1,-1,62,-1,-1,-1,-1,-1,18,-1,-1,-1,36,57,-1,-1,-1,15,-1,-1,-1,-1,29,-1,-1,-1,-1,
    46,-1,-1,-1,-1,6,-1,-1,34,-1,13,-1,-1,-1,32,-1,-1,-1,-1,-1,-1,50,-1,-1,-1,43,-1,-1,40,-1,-1,25,-1,-1,52,-1,-1,-1,
    61,-1,-1,-1,-1,-1,56,-1,-1,-1,28,-1,-1,45,-1,-1,-1,-1,12,-1,-1,-1,-1,49,-1,42,-1,-1,24,-1,-1,-1,-1,-1,-1,55,-1,27,
    -1,-1,-1,11,-1,-1,-1,-1,23,-1,-1,54,-1,-1,10,-1,-1,-1,-1,-1,9,-1,-1,-1,-1,-1,-1
};

/********************************
 * Helper function declarations *
 ********************************/
/// @brief Checks fast if the decoding is done from left to right
/// @var    g           probability grid containing the zero-mean probabilities
/// @var    startRow    index of first row of the subgrid to look at
/// @var    startCol    index of first column of the subgrid to look at
/// @return Number of votes in a region [-8,8] that claim the probability grid to decode from left to right
int forwardProbability(IntGrid g, int startRow, int startCol);

/// @brief Checks fast if the decoding is done from top to bottom
/// @var    g           probability grid containing the zero-mean probabilities
/// @var    startRow    index of first row of the subgrid to look at
/// @var    startCol    index of first column of the subgrid to look at
/// @return Number of votes in a region  [-8,8] that claim the probability grid to decode from top to bottom
int downwardProbability(IntGrid g, int startRow, int startCol);

///@brief adjusts the certainty of the rotation decoder
/// @var rot        Pointer to a rotationRecoder structure
/// @var rotProbX   number of votes in favor  - number of votes against having the right orientation in x
/// @var rotProbY   number of votes in favor  - number of votes against having the right orientation in y
void rotationDecoderUpdate(RotationDecoder* rot, int rotProbX, int rotProbY);

///@brief If rotationDecoder finds that the current rotation is false, change the axes and the origin of the grid coordinate system accordingly
///@var rot             Pointer to a rotationRecoder structure
///@var axes            Pointer to the mean grid directions (will be changed if the rotation decoder indicates a change)
///@var gridOffsetsInfo Pointer to a GridOffsetsInfo structure (will be changed if the rotation decoder indicates a change)
///@return the number of 90 degree rotations made by the function
int rotationDecoderUpdateAxes(RotationDecoder* rot, IntPoint* axes, GridOffsetsInfo* gridOffsetsInfo);

/// @brief      Rotate probability grids based on orientation of axes
/// @var        grids            probability grids to be rotated
/// @var        numRightTurns    number of right rotations that are needed to re-orient grid
void probabilityGridsTurn(ProbabilityGrids* grids, int numRightTurns);

/// @brief      Decodes the position given by probGrids
/// @var        probGrids   Are assumed to be in the right orientations
/// @var        startRow    Identified startRow to get the hightest cummulative probability
/// @var        startCol    Identified startCol to get the hightest probability
/// @return     Point with respective coordinates. If coordinates could not be determined the corresponding value in the point is set to -1
IntPoint decodePos(ProbabilityGrids probGrids, int startRow, int startCol, IntPoint* sector);

/// @brief  Determines the primary number of a row of the intgrid
/// @var    g           probability grid( negative values are interpreted as positive probabilities for a symbol to have value -1
/// @var    row         index of row to determine placement in mainnumbersequence
/// @var    startCol    index of the column where the decoding starts
/// @return placement of the row in main-number sequence (0-63) or INVALID_COEFF if placement is not determined
int getPrimaryNumberRow(IntGrid g, int row, int startCol);

/// @brief  Determines the primary number of a column of the intgrid
/// @var    g           probability grid( negative values are interpreted as positive probabilities for a symbol to have value -1
/// @var    startRow    index of the row where the decoding starts
/// @var    col         index of column to determine placement in mainnumbersequence
/// @return placement of the column in main-number sequence (0-63) or INVALID_COEFF if placement is not determined
int getPrimaryNumberCol(IntGrid g, int startRow, int col);

/// @brief   Correlates sequence with mainNumberSequence
/// @var     sequence The sequence
/// @var     length Number of entries in sequence
/// @return  Index of the peak
int correlationPeak(int* sequence, int length);

/// @brief Finds absolute position given a primary number sequence
/// @var primaryNumberSequence The sequence
/// @return Absolute position or -1 if the decoding failed
int decodePrimaryNumber(int* primaryNumberSequence);

///@brief Return-structure to give the 4 secondary numbers
typedef struct _Coefficients{
    char c[4];
} Coefficients;

///@brief Gets the four coefficients for a given shift difference
///@var dist Shift difference
Coefficients getCoeffs(int dist);

/// @brief  Searches for a pattern in the given sequence
/// @var    pattern         pointer to an array which is a part of a sequence
/// @var    sequence        pointer to a complete sequence with repetition of at least patternLength
/// @var    sequenceLength  period of the sequence
/// @var    patternLength   length of the pattern
/// @return Index of the first occurence of pattern in sequence or -1 if there is no occurence
int findSequenceOffset(const char* pattern, const unsigned char* sequence, unsigned char sequenceLength, unsigned char patternLength);

/// @brief Gets the position between 0 and 410815347 given the positions in secondary number sequences using the Chinese Remainder Theorem
/// @var a 4 positions in the 4 secondary number sequences
/// @return x such that x = a_i (mod length_secondary_num_seq_i) for all four i
int getPosition(const int* a);

/// @brief finds section of a pattern
/// @var    topLeftPosition decoded position of a pattern
/// @var    topMainShift    shift of the main number sequence of the top row
/// @var    leftMainShift   shift of the main number sequence of the left column
/// @return the X and the Y sector in an Intpoint structure
IntPoint findSection(IntPoint topLeftPosition, unsigned char topMainShift, unsigned char leftMainShift);

/// @brief Calculates the number of shifts of the main numbersequence occured at the given coordinate
/// @var position The coordinate
/// @return Number of shifts
unsigned int numShifts(unsigned int position);

/*********************************
 * Core function implementations *
 *********************************/
IntPoint probGridsDecode(ProbabilityGrids* probGrids, IntPoint pictureCenter, RotationDecoder* rotDec, IntPoint* axes, GridOffsetsInfo* gridOffsetsInfo, IntPoint* sector){

    int col = 0, row = 0;
    intGridFindBest8x8(&probGrids->maxProb, &col, &row);

    //find orientation:
    int fwdProb = forwardProbability(probGrids->probX, row, col);
    int dwdProb = downwardProbability(probGrids->probY, row, col);

    rotationDecoderUpdate(rotDec, dwdProb, fwdProb);

    int rotate = rotationDecoderUpdateAxes(rotDec, axes, gridOffsetsInfo);

    probabilityGridsTurn(probGrids, rotate);
    int tmp;
    switch(rotate%4){
    case 3:
        tmp = col;
        col = row;
        row = probGrids->probX.numRows - 8 - tmp;
        break;
    case 2:
        col = probGrids->probX.numCols - col - 8;
        row = probGrids->probX.numRows - row - 8;
        break;
    case 1:
        tmp = col;
        col = probGrids->probX.numCols - row - 8;
        row = tmp;
        break;
    }

    //find position:
    IntPoint pos = decodePos(*probGrids, row, col, sector);

    // Used to avoid having to use floats
    int divisorForResult = FIXED_RESULT_DIVISOR;
    // Handle subgrid accuracy
    if(pos.x >= 0)
        pos.x = (pos.x - row)*divisorForResult;
    if(pos.y >= 0)
        pos.y = (pos.y - col)*divisorForResult;
    int64_t deltaX = pictureCenter.x - gridOffsetsInfo->gridOrigin.x;
    int64_t deltaY = pictureCenter.y - gridOffsetsInfo->gridOrigin.y;
    int a = -axes[0].x;// Matrix elements of matrix
    int b = -axes[1].x;// a b
    int c = -axes[0].y;// c d
    int d = -axes[1].y;//
    int determinant = a*d - c*b;
    if(pos.x >= 0){
        pos.x -= (divisorForResult*(d*deltaX - b*deltaY))/determinant;
        pos.x += 4.5*divisorForResult;
    }
    if(pos.y >= 0){
        pos.y += 4.5*divisorForResult; //to get the grid center..
        pos.y -= (divisorForResult*(-c*deltaX + a*deltaY))/determinant;
    }

    return pos;
}

void rotationDecoderReset(RotationDecoder* rot){
    rot->x = 0;
    rot->y = 0;
}

void rotationDecoderDiminish(RotationDecoder* rot){
    rot->x /= 4;
    rot->y /= 4;
}

/***********************************
 * Helper function implementations *
 ***********************************/
int forwardProbability(IntGrid g, int startRow, int startCol){
    int i, j, lookup, probability, votes = 0;

    for(i=startCol; i<startCol+8; ++i){
        lookup = 0;
        probability = 1024;
        for(j=startRow; j<startRow+8; ++j){
            lookup |= (g.array[j][i] > 0) << (7 - (j - startRow));
            probability *= g.array[j][i];
            probability >>= 10;
        }
        probability = 1;

        if(mainLookUp[lookup] >= 0)
            votes+=probability;

        //reverse 8-bit lookup
        lookup = ((lookup * 0x0802LU & 0x22110LU) | (lookup * 0x8020LU & 0x88440LU)) * 0x10101LU >> 16;
        if(mainLookUp[(~lookup)&0xFF] >= 0)
            votes -= probability;

    }

    return votes;
}

int downwardProbability(IntGrid g, int startRow, int startCol){
    int i, j, lookup, probability, votes = 0;
    for(i=startRow; i<startRow+8; ++i){
        lookup = 0;
        probability = 1024;
        for(j=startCol; j<startCol+8; ++j){
            lookup |= (g.array[i][j] < 0) << (7 - (j - startCol));
            probability *= g.array[i][j];
            probability >>= 10;
        }
        probability = 1;

        if(mainLookUp[lookup] >= 0)
            votes += probability;

        //reverse 8-bit lookup
        lookup = ((lookup * 0x0802LU & 0x22110LU) | (lookup * 0x8020LU & 0x88440LU)) * 0x10101LU >> 16;
        if(mainLookUp[~(lookup)&0xFF] >= 0)
            votes -= probability;
    }

    return votes;
}

void rotationDecoderUpdate(RotationDecoder* rot, int rotProbX, int rotProbY){
    rot->x += rotProbX;
    rot->y += rotProbY;

    if(rot->x > ROTATION_DECODER_AWR)
        rot->x = ROTATION_DECODER_AWR;
    else if(rot->x < -ROTATION_DECODER_AWR)
        rot->x = -ROTATION_DECODER_AWR;

    if(rot->y > ROTATION_DECODER_AWR)
        rot->y = ROTATION_DECODER_AWR;
    else if(rot->y < -ROTATION_DECODER_AWR)
        rot->y = -ROTATION_DECODER_AWR;
}

int rotationDecoderUpdateAxes(RotationDecoder* rot, IntPoint* axes, GridOffsetsInfo* gridOffsetsInfo){
    int rotated = 0;

    if(rot->x < 0 && rot->y < 0){
        intPointMul(axes + 0, -1);
        intPointMul(axes + 1, -1);
        IntPoint tmpAxes = axes[0];
        axes[0] = axes[1];
        axes[1] = tmpAxes;
        rot->x *= -1;
        rot->y *= -1;

        //move origin coordinate
        gridOffsetsInfo->gridOrigin.x += (gridOffsetsInfo->gridMaxRows - 1)*axes[0].x + (gridOffsetsInfo->gridMaxCols - 1)*axes[1].x;
        gridOffsetsInfo->gridOrigin.y += (gridOffsetsInfo->gridMaxRows - 1)*axes[0].y + (gridOffsetsInfo->gridMaxCols - 1)*axes[1].y;

        rotated = 2;
    }
    else if(rot->x < 0 && rot->y > 0){
        IntPoint tmpAxes;
        tmpAxes = axes[0];
        axes[0].x = tmpAxes.y;
        axes[0].y = -tmpAxes.x;
        tmpAxes = axes[1];
        axes[1].x = tmpAxes.y;
        axes[1].y = -tmpAxes.x;

        rot->y = 0;
        rotated = 3;
        gridOffsetsInfo->gridOrigin.x -= (gridOffsetsInfo->gridMaxCols - 1)*axes[0].x;
        gridOffsetsInfo->gridOrigin.y -= (gridOffsetsInfo->gridMaxCols - 1)*axes[0].y;

    }
    else if(rot->y < 0 && rot->x > 0){
        IntPoint tmpAxes;
        tmpAxes = axes[0];
        axes[0].x = -tmpAxes.y;
        axes[0].y = tmpAxes.x;
        tmpAxes = axes[1];
        axes[1].x = -tmpAxes.y;
        axes[1].y = tmpAxes.x;

        gridOffsetsInfo->gridOrigin.x -= (gridOffsetsInfo->gridMaxRows - 1)*axes[1].x;
        gridOffsetsInfo->gridOrigin.y -= (gridOffsetsInfo->gridMaxRows - 1)*axes[1].y;
        rot->y = 0;
        rot->x = 0;
        rotated = 1;
    }

    return rotated;
}

void probabilityGridsTurn(ProbabilityGrids* grids, int numRightTurns){
    numRightTurns %= 4;

    // We can recycle the maxProb data arrays here since we've already found the location of max probability
    int* tmpGridList = grids->maxProb.data;
    int** tmpRowList = grids->maxProb.array;
    IntGrid tmpGrid;

    switch(numRightTurns){
    case 0:
        break;
    case 3:
        tmpGrid = intGridTurn(grids->probX, numRightTurns, -1, tmpGridList, tmpRowList);
        grids->probX = intGridTurn(grids->probY, numRightTurns, 1, (grids->probX).data, (grids->probX).array);
        intGridCopy(&grids->probY,&tmpGrid);
        break;
    case 2:
        tmpGrid = intGridTurn(grids->probX, numRightTurns, -1, tmpGridList, tmpRowList);
        intGridCopy(&grids->probX, &tmpGrid);
        tmpGrid = intGridTurn(grids->probY, numRightTurns, -1, tmpGridList, tmpRowList);
        intGridCopy(&grids->probY,&tmpGrid);
        break;
    case 1:
        tmpGrid = intGridTurn(grids->probX, numRightTurns, 1, tmpGridList, tmpRowList);
        grids->probX = intGridTurn(grids->probY, numRightTurns, -1, (grids->probX).data, (grids->probX).array);
        intGridCopy(&grids->probY,&tmpGrid);
        break;
    }
}

IntPoint decodePos(ProbabilityGrids probGrids, int startRow, int startCol, IntPoint* sector){
    IntPoint pos;
    int i;
    int primaryNumberSequence[DECODING_GRID_LENGTH];
    int leftMost;
    int topMost;

    for(i=DECODING_GRID_LENGTH; i--;){
        primaryNumberSequence[i] = getPrimaryNumberRow(probGrids.probY, startRow+i, startCol);
    }
    leftMost = primaryNumberSequence[0];
    pos.x = decodePrimaryNumber(primaryNumberSequence);

    for(i=DECODING_GRID_LENGTH; i--;){
        primaryNumberSequence[i] = getPrimaryNumberCol(probGrids.probX, startRow, startCol+i);
    }
    topMost = primaryNumberSequence[0];
    pos.y = decodePrimaryNumber(primaryNumberSequence);

    if(pos.x >= 0 && pos.y >= 0 && leftMost >= 0 && topMost >= 0 && sector){
        *sector = findSection(pos, leftMost, topMost);
    }

    return pos;
}

int getPrimaryNumberRow(IntGrid g, int row, int startCol){
    int primeRow, j, lookup=0, sequence[8];

    for(j=startCol; j<startCol+8; ++j){
        lookup |= (g.array[row][j] < 0) << (7 - (j - startCol));
        sequence[j - startCol] = -g.array[row][j];
    }
    primeRow = mainLookUp[lookup];

    if(primeRow == INVALID_COEFF)
        primeRow = correlationPeak(sequence,8);

    return primeRow;
}

int getPrimaryNumberCol(IntGrid g, int startRow, int col){
    int primeCol, j, lookup = 0, sequence[8];

    for(j=startRow; j<startRow+8; ++j){
        lookup |= (g.array[j][col] < 0) << (j - startRow);
        sequence[j - startRow] = g.array[j][col];
    }
    primeCol = mainLookUp[lookup];

    if(primeCol == INVALID_COEFF)
        primeCol = correlationPeak(sequence,8);

    return primeCol;
}

int correlationPeak(int* sequence, int length){
    int i, j;
    int result[MAIN_SEQUENCE_LENGTH];
    int maxResult = 0;
    int maxIndex = 0;

    for(i=MAIN_SEQUENCE_LENGTH; i--;){
        result[i] = 0;
        for(j=length; j--;){
            result[i] += sequence[j]*mainNumberSequenceForCorr[i+j];
        }
        if(result[i] > maxResult){
            maxResult=result[i];
            maxIndex=i;
        }
    }

    return maxIndex;
}

int decodePrimaryNumber(int* primaryNumberSequence){
    unsigned char patternLength = DECODING_GRID_LENGTH-1;
    unsigned char i;
    int places[4];
    int secondaryNumberSequence[patternLength];
    char cPatterns[4][patternLength];

    for(i=patternLength; i--;){
        secondaryNumberSequence[i] = primaryNumberSequence[i+1] - primaryNumberSequence[i];
        secondaryNumberSequence[i] += (secondaryNumberSequence[i] < 0)*MAIN_SEQUENCE_LENGTH;
    }

    for(i=patternLength; i--;){
        if(secondaryNumberSequence[i] < 5 || secondaryNumberSequence[i] > 58){
            return -1;
        }

        //split into coefficients.
        Coefficients tmp = getCoeffs(secondaryNumberSequence[i]);
        cPatterns[0][i] = tmp.c[0];
        cPatterns[1][i] = tmp.c[1];
        cPatterns[2][i] = tmp.c[2];
        cPatterns[3][i] = tmp.c[3];
    }

    //lookup coefficents place in sequences;
    for(i=4; i--;){
        places[i] = findSequenceOffset(cPatterns[i], cSequences[i], cSequenceLengths[i], patternLength);
        if(places[i] < 0){
            return -1;
        }
    }

    int position = getPosition(places);

    return position;
}

Coefficients getCoeffs(int dist){
    Coefficients coeff;

    dist -= 5;
    coeff.c[3] = dist/18;
    dist -= 18*coeff.c[3];
    coeff.c[2] = dist/9;
    dist -= 9*coeff.c[2];
    coeff.c[1] = dist/3;
    dist -= 3*coeff.c[1];
    coeff.c[0] = dist;

    return coeff;
}

int findSequenceOffset(const char* pattern, const unsigned char* sequence, unsigned char sequenceLength, unsigned char patternLength){
    int i, j;

    for(i=0; i<sequenceLength; ++i){
        for(j=0; j<patternLength; ++j){
            if(pattern[j] != sequence[j+i])
                break;
        }
        if(j == patternLength)
            return i;
    }

    return -1;
}

int getPosition(const int* a){
    long long x = 0;

    int i;
    for(i=4;i--;)
        x += (long long)a[i]*cSequenceLengthsMulInvs[i]*otherCSequenceLengthsProduct[i];

    return x%cSequenceLengthsProduct;
}

IntPoint findSection(IntPoint topLeftPosition, unsigned char topMainShift, unsigned char leftMainShift){
    IntPoint section;

    section.x = topMainShift - numShifts(topLeftPosition.x) - topLeftPosition.y%63;
    section.y = leftMainShift - numShifts(topLeftPosition.y) - topLeftPosition.x%63;
    if(section.x < 0)
        section.x += 63;
    if(section.y < 0)
        section.y += 63;

    return section;
}

unsigned int numShifts(unsigned int position){
    unsigned int numshifts = 5*(position%MAIN_SEQUENCE_LENGTH);
    unsigned char i, j, rest;
    for(i=0; i<4; i++){
        rest = position%cSequenceLengths[i];
        numshifts += (cSequenceMultipliers[i]*cSequenceShifts[i]*(position/cSequenceLengths[i]))%MAIN_SEQUENCE_LENGTH;
        for(j=0; j<rest; j++){
            numshifts += cSequenceMultipliers[i]*cSequences[i][j];
        }
    }
    return numshifts%MAIN_SEQUENCE_LENGTH;
}

#pragma GCC pop_options

#ifdef __cplusplus
}
#endif
