/**
 * @file dots_localization.c
 * @brief Source of the localization functionality
 * @author Lukas Hostettler
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#include "compiler.h"
#include "config.h"
#include "decoder.h"
#include "grid.h"
#include "gridprobability.h"
#include "imgproc.h"

#include "dots_localization.h"

#pragma GCC push_options
#pragma GCC optimize ("O1")

static DotsPositionInfo dotPosInfo;
DotsPositionInfo dotsPosInfoInit();

DotsPositionInfo* dotsLocalize(DotsPositionInfo* previousInfo, unsigned char* image, int rows, int cols){
    /*
     * Initial setup
     */
    if(!image || rows < 1 || cols < 1)
        return 0;
    if(!previousInfo) {
        dotPosInfo = dotsPosInfoInit();
        previousInfo = &dotPosInfo;
    }

    previousInfo->decodingError = DECODING_NO_ERROR;
    previousInfo->decoded = 0;

    /*
     * Image processing
     */

    //Blob detection
    int thresholdValue = imageMean(image, rows, cols) - THRESHOLD_OFFSET;
    ImgSegmentList segList = detectPoints(image, rows, cols, thresholdValue);
    int numPoints = segList.numElements;

    if(numPoints < 64){
        previousInfo->decodingError = DECODING_NOT_ENOUGH_POINTS;
        return previousInfo;
    }
    if(numPoints >= MAX_NUM_POINTS){
        previousInfo->decodingError = DECODING_TOO_MANY_POINTS;
        return previousInfo;
    }

    //Sort points based on y values
    IntPoint* pointArray = sortPoints(segList);

    //Distortion correction
    undistortPoints(pointArray, numPoints, rows, cols, CAMERA_DISTORTION_PARAMETER);

    /*
     * Grid estimation
     */
    //Find initial grid
    IntPoint origin;
    findInitialGridEstimate(pointArray, numPoints, previousInfo->axes, &origin);

    //Create final grid estimates
    int correctGridSuccess = correctGridEstimate(pointArray, numPoints, previousInfo->axes, &origin);
    if (!correctGridSuccess) {
        previousInfo->decodingError = DECODING_COLLINEAR_GRID;
        rotationDecoderDiminish(&previousInfo->rotDec);
        return previousInfo;
    }

    /*
     * Probabilistic treatment
     */
    //Assign probabilities
    GridOffsetsInfo gridOffsetsInfo = calculateGridOffsets(pointArray, numPoints, previousInfo->axes, &origin);
    if(gridOffsetsInfo.gridMaxCols >= MAX_GRID_LENGTH || gridOffsetsInfo.gridMaxRows >= MAX_GRID_LENGTH){
        previousInfo->decodingError = DECODING_WRONG_GRID_SIZE;
        return previousInfo;
    }

    ProbabilityGrids probGrids = makeProbabilityGrids(gridOffsetsInfo);
    if(probGrids.maxProb.numCols < 8 || probGrids.maxProb.numRows < 8) {
        previousInfo->decodingError = DECODING_BOTH_FAILED;
        return previousInfo;
    }

    /*
     * Decoding of probability grids
     */
    IntPoint pictureCenter = {(rows*SUBDIVISION/2), (cols*SUBDIVISION/2)};
    IntPoint decodedPos = probGridsDecode(&probGrids, pictureCenter, &previousInfo->rotDec, previousInfo->axes, &gridOffsetsInfo, &previousInfo->sector);

    if(decodedPos.x >= 0)
        previousInfo->position.x = decodedPos.x;
    else
        previousInfo->decodingError = DECODING_X_FAILED;

    if(decodedPos.y >= 0)
        previousInfo->position.y = decodedPos.y;
    else
        previousInfo->decodingError = DECODING_Y_FAILED;

    if(decodedPos.x < 0 && decodedPos.y < 0)
        previousInfo->decodingError = DECODING_BOTH_FAILED;

    previousInfo->decoded = (previousInfo->decodingError == DECODING_NO_ERROR);
    return previousInfo;
}

DotsPositionInfo dotsPosInfoInit(){
    DotsPositionInfo posInfo;

    initAxes(posInfo.axes, AXIS_INIT_LENGTH);
    rotationDecoderReset(&posInfo.rotDec);
    posInfo.decoded = 0;
    posInfo.decodingError = DECODING_NO_ERROR;
    return posInfo;
}

float dotsGetFloatX(DotsPositionInfo* info){
    if(info)
        return info->position.x/(float)info->divisor;
    return -1;
}

float dotsGetFloatY(DotsPositionInfo* info){
    if(info)
        return info->position.y/(float)info->divisor;
    return -1;
}

int dotsGetIntX(DotsPositionInfo* info){
    if(info)
        return info->position.x;
    return -1;
}

int dotsGetIntY(DotsPositionInfo* info){
    if(info)
        return info->position.y;
    return -1;
}

int dotsGetSectorX(DotsPositionInfo* info){
    if(info)
        return info->sector.x;
    return -1;
}

int dotsGetSectorY(DotsPositionInfo * info){
    if(info)
        return info->sector.y;
    return -1;
}

float dotsGetAngle(DotsPositionInfo* info){
    if(info)
        return atan2(info->axes[0].y, info->axes[0].x);
    return 0;
}

int dotsGetDivisor(DotsPositionInfo* info){
    if(info)
        return info->divisor;
    return -1;
}

void dotsSetDivisor(DotsPositionInfo* info, int divisor){
    if(info && divisor)
        info->divisor = divisor;
}


int dotsIsDecoded(DotsPositionInfo* info){
    if(info)
        return info->decoded;
    return 0;
}

DecodingError dotsGetError(DotsPositionInfo* info){
    if(info)
        return info->decodingError;
    return NO_INFO;
}

int dotsIsAboveGround(DotsPositionInfo* info){
    if(info){
        switch(info->decodingError){
            case DECODING_NOT_ENOUGH_POINTS:
            case DECODING_COLLINEAR_GRID:
                return 1;
            default:
                return 0;
        }
    }
    return 1;
}

#pragma GCC pop_options
