/**
 * @file intgrid.h
 * @brief IntGrid strcture header
 * @author Lukas Hostettler
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#ifndef INTGRID_H
#define INTGRID_H

#ifdef __cplusplus
extern "C" {
#endif

#include "config.h"
#include "intlist.h"

/// Symbol for a missing point
#define NOPOINT -1

typedef struct _IntGrid{
    int numCols;
    int numRows;
    int* data;   //!<The actual pointer to data.
    int** array; //! Pointers construct for easy access: IntGrid[numRow][numCol];
} IntGrid; //! a dynamic array


///@brief   Creates an empty 0x0 IntGrid. There is no data allocation.
///@return  empty IntGrid Structure
IntGrid intGridCreateEmpty();

///@brief   Creates an empty IntGrid of size numCol x numRow.
///@var     numCol      number of columns of the grid
///@var     numRow      nomber of rows of the grid
///@var     gridList    pointer to integer array to use to store data
///@var     rowList     pointer to an array of pointers to the start ofeach of the rows of the grid
///@return  an allocated IntGrid of size numCol x numRow
IntGrid intGridCreate(int numCol, int numRow, int* gridList, int** rowList);

/// @brief Populate an intgrid with a value.
/// @var    g   IntGrid structure
/// @var    value  The integer value to fill g with
void intGridFill(IntGrid *g, int value);

/// @brief Copies grid data from one to another
/// @var destination   destination IntGrid structure
/// @var source        source IntGrid structure
void intGridCopy(IntGrid* destination, IntGrid* source);

///@brief   Find the largest value in the grid
///@var     g           the IntGrid in question
///@var     idx         the pointer where we will store the max value in the grid
int intGridMaximum(IntGrid *g, int* idx);

///@brief   Convert from an index into row,col pair
///@var     g           the IntGrid in question
///@var     col      the resulting column
///@var     row      the resulting row
void intGridIndToSub(IntGrid *g, int ind, int* col, int* row);

///@brief   Turns the intGrid
///@var     grid              the Intgrid to be turned
///@var     numRightTurns     number of turns to be done
///@var     factor            multiplicative factor applied to all entries (for inversing the probabilities)
///@var     gridList,rowList  pass in the statically allocated data arrays to be supplied to the returned IntGrid
///@return  A turned version of the given grid
IntGrid intGridTurn(IntGrid grid, int numRightTurns, int factor, int* gridList, int** rowList);

/// @brief Finds the closest existing element larger than zero in an intGrid
/// @var    g   IntGrid structure
/// @var    x   Pointer for returning the x index of the found element in g, also serves as initial input of the spiral
/// @var    y   Pointer for returning the y index of the found element in g, also serves as initial input of the spiral
/// @return 1 for success, 0 for failure
int intGridFindClosestTo(IntGrid g, int* x, int* y);

///@brief   finds index of top,left corner of the sub square of size n x n with the hightest sum of values
///@var     grid        IntGrid to be examined
///@var     col        pointer to an integer. Recieves leftmost column index of the highest sum of values square
///@var     row        pointer to an integer. Recieves topmost row index of the highest sum of values square
///@var     n           sidelength of the subsquare
void intGridFindBest8x8(IntGrid* grid, int* col, int* row);

///@brief   makes a convolution with a square window of size windowsize and coefficient 1
///@var     g           the IntGrid to be convoluted
///@return  An Intgrid containing the entries of the inner convolution. Its size is g.numRows-windowSize+1 x g.numCols-windowsSize+1
IntGrid intGridLocalSum(IntGrid *g);

#ifdef __cplusplus
}
#endif

#endif //INTGRID_H

