/**
 * @file intpoint.h
 * @brief IntPoint structure header
 * @author Lukas Hostettler
 * @version 2.0
 * @date 2015-15-12
 */

#ifndef INTPOINT_H
#define INTPOINT_H

#ifdef __cplusplus
extern "C" {
#endif

#include <math.h>

#pragma GCC push_options
#pragma GCC optimize ("O1")

///@brief Describes a 2D point with integer coordinates
typedef struct _IntPoint{
    int x;
    int y;
} IntPoint;

///@brief Stores the certainty of the rotation estimate
typedef IntPoint RotationDecoder;

///@brief Adds (x,y) to given point
///@var x Value to add
///@var y Value to add
///@var p Point to add to
static inline void intPointAdd(IntPoint* p, int x, int y){
    p->x += x;
    p->y += y;
}

///@brief Multiplies the given point with a scalar
///@var p Point to multiply
///@var factor The scalar factor
static inline void intPointMul(IntPoint* p, int factor){
    p->x *= factor;
    p->y *= factor;
}

///@brief Gets the euclidean norm of the point as if it were a vector
///@var p point of interest
///@return Euclidean norm
static inline IntPoint intPointNormalize(IntPoint p){
    IntPoint normalized = p;
    unsigned int sqsum = p.x*p.x+p.y*p.y;
    if (sqsum) {
        // multiply by 100000 for integer precision
        float multiplier = 100000/sqrt(sqsum);
        normalized.x *= multiplier;
        normalized.y *= multiplier;
    }
    return normalized;
}

///@brief Gets the difference between two points
///@var p1 First point
///@var p2 Second point
///@return p1 - p2
static inline IntPoint intPointDiff(IntPoint p1, IntPoint p2){
    IntPoint diff = {p1.x - p2.x, p1.y - p2.y};
    return diff;
}

#pragma GCC pop_options

#ifdef __cplusplus
}
#endif

#endif //INTPOINT_H
