/**
 * @file webcam_sample.c
 * @brief Source for the sample webcam image acquisition and calling libdots
 * @author Lukas Hostettler
 * @author Ayberk Özgür
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#include <iomanip>
#include <iostream>
#include <stdio.h>
#include <stdarg.h>
#include <time.h>

#include "webcam_sample.h"

#define GRID_SIZE 0.508                         ///@var Average distance in mm between two adjacent dots, in this example
#define CUSTOM_THRESHOLD_OFFSET 40              ///@var This test webcam needs more offset
#define CUSTOM_CAMERA_DISTORTION_PARAMETER 0    ///@var This test webcam does not need dist. correction
#define IMG_HEIGHT 240                          ///@var Height of grabbed frame
#define IMG_WIDTH 320                           ///@var Width of grabbed frame
#define USE_STATICIMAGE 0                       ///@var 1 - Use a static image to test, 0 - Use input from a camera

using namespace std;

WebcamSample::WebcamSample(int camNo){
    cap = VideoCapture(camNo);
    if(!cap.isOpened()){
        cout << "Cannot open the video cam" << endl;
        return;
    }
    cap.set(CV_CAP_PROP_FRAME_HEIGHT, IMG_HEIGHT);
    cap.set(CV_CAP_PROP_FRAME_WIDTH, IMG_WIDTH);

    //get the width of frames of the video
    double dWidth = cap.get(CV_CAP_PROP_FRAME_WIDTH);
    //get the height of frames of the video
    double dHeight = cap.get(CV_CAP_PROP_FRAME_HEIGHT);

    cout << "Frame size : " << dWidth << "x" << dHeight << endl;

    dotPosInfo = dotsPosInfoInit();
}

WebcamSample::~WebcamSample(){
    cap.release();
}

bool WebcamSample::run(){
    if(waitKey(10) != -1)
        return false;

    Mat frame;
    if (!getRawFrame(frame)) {
        //If unable to find frame, skip this run loop
        cout << "Could not get frame!" << endl;
        return true;
    }

    vector<Mat> channels;
    split(frame, channels);
    Mat originalFrame = frame.clone();
    unsigned char* image = channels[1].ptr();
    int rows = channels[1].rows;
    int cols = channels[1].cols;

    //We do not use dotsLocalize() in this sample to show each step of the pipeline
    DotsPositionInfo *posInfo = &dotPosInfo;
    posInfo->decodingError = DECODING_NO_ERROR;
    posInfo->decoded = 0;

    /*
     * Image processing
     */

    //Blob detection
    int thresholdValue = imageMean(image, rows, cols) - CUSTOM_THRESHOLD_OFFSET;
    ImgSegmentList segList = detectPoints(image, rows, cols, thresholdValue);

    int numPoints = segList.numElements;
    if(numPoints < 64){
        posInfo->decodingError = DECODING_NOT_ENOUGH_POINTS;
        imshow("Original Image (assuming mirrored)", originalFrame);
        printInfo();
        return true;
    }
    if(numPoints >= MAX_NUM_POINTS){
        posInfo->decodingError = DECODING_TOO_MANY_POINTS;
        imshow("Original Image (assuming mirrored)", originalFrame);
        printInfo();
        return true;
    }

    IntPoint* pointArray = sortPoints(segList);

    //Distortion correction
    undistortPoints(pointArray, numPoints, rows, cols, CUSTOM_CAMERA_DISTORTION_PARAMETER);

    /*
     * Grid estimation
     */

    //Find initial grid
    IntPoint origin;
    findInitialGridEstimate(pointArray, numPoints, posInfo->axes, &origin);

    //Explicitly find neighbours again for visualization purposes
    Neighbours* neigh = findNeighbours(pointArray, numPoints);
    drawNeighbourhood(frame, pointArray, neigh, numPoints);

    //Refine grid estimate
    int correctGridSuccess = correctGridEstimate(pointArray, numPoints, posInfo->axes, &origin);

    if(!correctGridSuccess) {
        posInfo->decodingError = DECODING_COLLINEAR_GRID;
        rotationDecoderDiminish(&posInfo->rotDec);
        displayLocalizationResults(originalFrame, frame);
        return true;
    }

    /*
     * Probabilistic treatment
     */

    //Assign probabilities
    GridOffsetsInfo gridOffsetsInfo = calculateGridOffsets(pointArray, numPoints, posInfo->axes, &origin);

    if(gridOffsetsInfo.gridMaxCols >= MAX_GRID_LENGTH || gridOffsetsInfo.gridMaxRows >= MAX_GRID_LENGTH){
        posInfo->decodingError = DECODING_WRONG_GRID_SIZE;
        displayLocalizationResults(originalFrame, frame);
        return true;
    }

    ProbabilityGrids probGrids = makeProbabilityGrids(gridOffsetsInfo);
    //Print probabilities and assigned offset directions
    drawPointProbabilities(frame, probGrids.maxProb, posInfo->axes, gridOffsetsInfo.gridOrigin);
    drawPointClassification(frame, probGrids, posInfo->axes, gridOffsetsInfo.gridOrigin);

    if(probGrids.maxProb.numCols < 8 || probGrids.maxProb.numRows < 8) {
        posInfo->decodingError = DECODING_BOTH_FAILED;
        displayLocalizationResults(originalFrame, frame);
        return true;
    }

    /*
     * Decoding of probability grids
     */

    IntPoint pictureCenter = {(rows*SUBDIVISION/2), (cols*SUBDIVISION/2)};
    IntPoint decodedPos = probGridsDecode(&probGrids, pictureCenter, &posInfo->rotDec, posInfo->axes, &gridOffsetsInfo, &posInfo->sector);

    //Find the best area to decode separately again just to draw on screen
    int col = 0, row = 0;
    intGridFindBest8x8(&probGrids.maxProb, &col, &row);
    drawDecodingFrame(frame, row, col, posInfo->axes, gridOffsetsInfo.gridOrigin);

    if(decodedPos.x >= 0)
        posInfo->position.x = decodedPos.x;
    else
        posInfo->decodingError = DECODING_X_FAILED;

    if(decodedPos.y >= 0)
        posInfo->position.y = decodedPos.y;
    else
        posInfo->decodingError = DECODING_Y_FAILED;

    if(decodedPos.x < 0 && decodedPos.y < 0)
        posInfo->decodingError = DECODING_BOTH_FAILED;

    posInfo->decoded = (posInfo->decodingError == DECODING_NO_ERROR);
    displayLocalizationResults(originalFrame, frame);
    return true;
}

bool WebcamSample::getRawFrame(Mat& frame){
#if USE_STATICIMAGE
    frame = imread("testImage.png", CV_LOAD_IMAGE_COLOR);
    return !!frame.data;
#else
    bool ans = false;
    if(cap.isOpened())
        ans = cap.read(frame);
    return ans;
#endif
}

bool WebcamSample::getRawFrame(Mat& frame, Rect roi){
    bool ans = false;
    Mat tmp;
    if(cap.isOpened()){
        ans = cap.read(tmp);
        frame = tmp(roi).clone();
    }
    return ans;
}

void WebcamSample::printInfo(){
    float robotX = ((float)dotsGetIntX(&dotPosInfo))/FIXED_RESULT_DIVISOR*GRID_SIZE;
    float robotY = ((float)dotsGetIntY(&dotPosInfo))/FIXED_RESULT_DIVISOR*GRID_SIZE;
    float robotTheta = dotsGetAngle(&dotPosInfo)/3.14159*180;

    cout << "x: " << robotX << " mm, y:" << robotY << " mm, theta: " << robotTheta << " degrees";

    switch(dotPosInfo.decodingError){
        case DECODING_NO_ERROR:
            cout << endl;
            break;
        case DECODING_TOO_MANY_POINTS:
            cout << "; Decoding error: Too many points" << endl;
            break;
        case DECODING_NOT_ENOUGH_POINTS:
            cout << "; Decoding error: Not enough points" << endl;
            break;
        case DECODING_COLLINEAR_GRID:
            cout << "; Decoding error: Collinear grid" << endl;
            break;
        case DECODING_WRONG_GRID_SIZE:
            cout << "; Decoding error: Wrong grid size" << endl;
            break;
        case DECODING_X_FAILED:
            cout << "; Decoding error: Decoding X failed" << endl;
            break;
        case DECODING_Y_FAILED:
            cout << "; Decoding error: Decoding Y failed" << endl;
            break;
        case DECODING_BOTH_FAILED:
            cout << "; Decoding error: Decoding both X and Y failed" << endl;
            break;
        case NO_INFO:
            cout << "; Decoding error: No info" << endl;
            break;
    }
}

void WebcamSample::displayLocalizationResults(Mat& originalFrame, Mat& frame) {
    printInfo();
    drawRobot();
    imshow("Original Image (assuming mirrored)", originalFrame);
    imshow("Processed Image (press any key to exit)", frame);
}

void WebcamSample::drawRobot(int width, int height, int radius){
    static Scalar background(255,255,255);
    static Scalar black(0,0,0);

    Mat I(height, width, CV_8UC3, background);

    float robotX = (float)dotsGetIntX(&dotPosInfo)/FIXED_RESULT_DIVISOR*GRID_SIZE;
    float robotY = (float)dotsGetIntY(&dotPosInfo)/FIXED_RESULT_DIVISOR*GRID_SIZE;
    float robotTheta = dotsGetAngle(&dotPosInfo);

    if(robotX >= 0 && robotY >= 0){
        Point pos(robotX, robotY);
        Point dir(radius*cosf(robotTheta), radius*sinf(-robotTheta));
        circle(I, pos, radius, black);
        line(I, pos, pos + dir, black);
    }

    imshow("Robot on A4 Sheet", I);
}

void WebcamSample::drawNeighbourhood(Mat& I, IntPoint* pointArray, Neighbours* neigh, int numPoints, Scalar color){
    //We ignored neighbourhood relations for the 1st and last point for the sake of optimization and code cleanliness so don't draw them here
    int subdivision = SUBDIVISION;
    for(int i=1;i<numPoints-1;i++){
        Point p1(pointArray[i].x/subdivision, pointArray[i].y/subdivision);
        Neighbours nei = neigh[i];
        for(int j=0;j<4;j++){
            Point p2(pointArray[nei.neigbourIdx[j]].x/subdivision, pointArray[nei.neigbourIdx[j]].y/subdivision);
            line(I, p1, p2, color);
        }
    }
}

void WebcamSample::drawPointProbabilities(Mat& I, IntGrid g, IntPoint* axes, IntPoint origin, int min, int max){
    int row, col;
    for(row=0;row<g.numRows;++row){
        for(col=0;col<g.numCols;++col){
            Point center = pointFromIndex(col, row, axes, origin);
            int colorVal = (255*g.array[row][col])/(max - min);
            circle(I, center, 5, Scalar(0, colorVal, 255 - colorVal), -1);
        }
    }
}

void WebcamSample::drawPointClassification(Mat& I, ProbabilityGrids grids, IntPoint* axes, IntPoint origin){
    int row, col;
    string text;
    int fontFace = CV_FONT_HERSHEY_PLAIN;
    double fontScale = 0.5;

    origin.x -= 300;
    origin.y += 300;

    for(row=0;row<grids.maxProb.numRows;++row){
        for(col=0;col<grids.maxProb.numCols;++col){
            Point center = pointFromIndex(col, row, axes, origin);
            if(grids.probX.array[row][col] < 0 && grids.probY.array[row][col] < 0)
                text = "L";
            else if(grids.probX.array[row][col] > 0 && grids.probY.array[row][col] < 0)
                text = "D";
            else if(grids.probX.array[row][col] < 0 && grids.probY.array[row][col] > 0)
                text = "U";
            else if(grids.probX.array[row][col] > 0 && grids.probY.array[row][col] > 0)
                text = "R";
            putText(I, text, center, fontFace, fontScale, Scalar(0,0,0));
        }
    }
}

void WebcamSample::drawDecodingFrame(Mat& I, int row, int col, IntPoint* axes, IntPoint origin, Scalar color){
    Point pt0 = pointFromIndex(col, row, axes, origin);
    Point pt1 = pointFromIndex(col + 7, row, axes, origin);
    Point pt2 = pointFromIndex(col + 7, row + 7, axes, origin);
    Point pt3 = pointFromIndex(col, row + 7, axes, origin);

    circle(I, pt0, 10, color, 2);
    line(I, pt0, pt1, color, 2);
    line(I, pt1, pt2, color, 2);
    line(I, pt2, pt3, color, 2);
    line(I, pt3, pt0, color, 2);
}

Point WebcamSample::pointFromIndex(int col, int row, IntPoint* axes, IntPoint origin){
    int px = (origin.x + row*axes[0].x + col*axes[1].x)/SUBDIVISION;
    int py = (origin.y + row*axes[0].y + col*axes[1].y)/SUBDIVISION;
    return Point(px, py);
}
