/**
 * @file webcam_sample.h
 * @brief Header for the sample webcam image acquisition and calling libdots
 * @author Lukas Hostettler
 * @author Ayberk Özgür
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#ifndef WEBCAM_SAMPLE_H
#define WEBCAM_SAMPLE_H

#include <opencv2/opencv.hpp>

#include "compiler.h"
#include "config.h"
#include "decoder.h"
#include "dots_localization.h"
#include "imgproc.h"
#include "grid.h"
#include "gridprobability.h"

using namespace cv;

class WebcamSample {

public:

    /// @brief  Constructor, establishing connection to a webcam
    /// @var    camNo   number of the camera to connect
    /// @return a WebcamSample class
    WebcamSample(int camNo = 0);

    /// @brief  Destructor, closes connection to webcam. Frees the memory
    ~WebcamSample();

    /// @brief Runs this samples
    /// @return Whether to keep running, or a quit signal was received
    bool run();

private:

    /// @brief  grabs a frame from the camera and stores it in frame
    /// @var    frame   reference to a cv::Mat class
    /// @return success of operation
    bool getRawFrame(Mat& frame);

    /// @brief  grabs a frame from the camera and stores it in frame
    /// @var    frame   reference to a cv::Mat class
    /// @var    roi     Region of interest, only the region of interest will be treated
    /// @return success of operation
    bool getRawFrame(Mat& frame, Rect roi);

    /// @brief Prints pose and error if present
    void printInfo();

    /// @brief Displays the results of the localization attempt
    void displayLocalizationResults(Mat& originalFrame, Mat& frame);

    /// @brief  opens a window if not yet opened and draws the current robot position
    /// @var    width   width of the coded paper in mm
    /// @var    heigh   height of the coded paper in mm
    /// @var    radius  radius of the device
    void drawRobot(int width = 210, int height = 297, int radius = 30);

    /// @brief Draws the neighbourhood of our grid by drawing lines between points that are mutual neighbours
    void drawNeighbourhood(Mat& I, IntPoint* pointArray, Neighbours* neigh, int numPoints, Scalar color = Scalar(0,0,0));

    /// @brief Draws the detected points over the original frame with colors that reflect the corresponding grid probabilities
    void drawPointProbabilities(Mat& I, IntGrid g, IntPoint* axes, IntPoint origin, int min = 0, int max = 1024*256);

    /// @brief Draws the classification of each point offset (UP, DOWN, LEFT or RIGHT) on top of the original frame
    void drawPointClassification(Mat& I, ProbabilityGrids grids, IntPoint* axes, IntPoint origin);

    /// @brief Draws the frame around the 8x8 grid space that we are using to decode the position
    void drawDecodingFrame(Mat& I, int row, int col, IntPoint* axes, IntPoint origin, Scalar color = Scalar(255,0,0));

    /// @brief Helper method to convert a row, col index to a OpenCV Point structure
    Point pointFromIndex(int col, int row, IntPoint* axes, IntPoint origin);

    /// @brief  structure for the webcam handling
    cv::VideoCapture cap;

    /// @brief  pointer to structure containing all information relating to the decoding
    DotsPositionInfo dotPosInfo;
};

#endif // WEBCAM_SAMPLE_H

